/* TEXTURE.CPP - Programmed by Jonathan Hobson */
/* 100404: Maarten Hofman: Allow reading of the name */
/* 040704: Maarten Hofman: Added transparancy */
/* 090704: Maarten Hofman: Added animation */
/* 120704: Maarten Hofman: Added debug output */
/* 280704: Maarten Hofman: Fixed EOP=1 problem */

#include <png.h>
#include <string.h>
#include "texture.h"
#include "dma.h"

CTexture::CTexture() {
  trans=false;
  next=NULL;
  timing=0;
  speed=0;
}


CTexture::~CTexture()
{

}

CTexture *CTexture::get_next() {
  timing++;
  if(timing<speed) return NULL;
  timing=0;
  return next;
}
 
void CTexture::set_next(CTexture *what,int how) {
  next=what;
  speed=how;
}

bool CTexture::is_trans() {
  return trans;
}

void CTexture::set_trans() {
  trans=true;
}

uint32 TextureLog2(uint32 n)
{
	switch(n)
	{
	case 256:
		return 8;
	case 128:
		return 7;
	case 64:
		return 6;
	case 32:
		return 5;
	default:
		printf("Illegal texture size.\n");
		exit(-1);
	}
}

uint32 TSMFromColourDepth(uint32 depth)
{
	switch(depth)
	{
	case 32:
		return PSMCT32;
	case 8:
		return PSMT8;
	default:
		printf("Unsupported colour depth.\n");
		exit(-1);
	}
}

int CTexture::has_name(char *how) {
  return !strcmp(name,how);
}

void CTexture::set_name(char *how) {
  strncpy(name,how,32);
}

bool CTexture::LoadImage(const char * strFilename, bool bRedAsAlpha)
{
	static const int bpptbl[] = { 16, 24, 32, 4, 8 };

	FILE * fp = fopen(strFilename, "rb");

	if(fp == NULL)
	{
		fprintf(stderr,"Failed to load %s\n", strFilename);
		return false;
	}

        strncpy(name,strFilename,32);

	fseek(fp, 0, SEEK_END );
	unsigned int uiFileSize = ftell(fp);
	fseek(fp, 0, SEEK_SET );

	if(uiFileSize == 0)
	{
		printf("%s is zero size.\n", strFilename);
		return false;
	}

	unsigned char *pTmpData = new unsigned char[uiFileSize];
	fread(pTmpData, 1, uiFileSize, fp);
	fclose(fp);

	if(!Tim2CheckFileHeader(pTmpData))
	{
		printf("Tim2CheckFileHeader failed on %s.\n", strFilename);
		return false;
	}

	TIM2_FILEHEADER *fh = (TIM2_FILEHEADER*)pTmpData;
	TIM2_PICTUREHEADER *ph;
	ph = Tim2GetPictureHeader(pTmpData, 0);

	int bpp = bpptbl[ph->ImageType - 1];
	
	m_iWidth = ph->ImageWidth;
	m_iHeight = ph->ImageHeight;
	m_iImageOriginalDepth = m_iImageDepth = bpp;
	m_iDataSize = ph->ImageWidth * ph->ImageHeight * bpp / 8;
	
	if(fh->FormatVersion == 0xEE)
	{
		char *ptrUserData = ((char*)ph) + sizeof(TIM2_PICTUREHEADER);
		m_l2Width = TextureLog2(*(short*)ptrUserData);
		m_l2Height = TextureLog2(*(short*)(ptrUserData+2));
		switch (*(ptrUserData+4))
		{
			case TIM2_IDTEX4: m_iImageOriginalDepth = 4;  break;
			case TIM2_IDTEX8: m_iImageOriginalDepth = 8;  break;
			case TIM2_RGB16:  m_iImageOriginalDepth = 16; break;
			case TIM2_RGB24:  m_iImageOriginalDepth = 24; break;
			case TIM2_RGB32:  m_iImageOriginalDepth = 32; break;
		}
	}

	m_iTexAddr = GIFStaticDMA.GetPointer();

	assert((m_iDataSize & 0xF) == 0);	// Make sure the texture data is a quadword multiple

	uint32 *pImage = (uint32 *)((char *)ph + ph->HeaderSize);

	for(int iQW = 0; iQW < (m_iDataSize >> 2); iQW++)
	{
		union
		{
			unsigned int uiValue;
			unsigned char cValue[4];
		} Colour;

		Colour.uiValue = *(pImage++);

		// TODO: What about 16 bit images?
		if(m_iImageOriginalDepth > 8)
		{
			uint8 red = Colour.cValue[0];
			Colour.cValue[0] = Colour.cValue[2];
			Colour.cValue[2] = red;
		}

		GIFStaticDMA.Add32(Colour.uiValue);
	}

	GIFStaticDMA.DMARet();

	int iNbCols;

	if ( m_iImageOriginalDepth <= 8 )
	{
		unsigned char *pClut = (unsigned char *)((char *)ph + ph->HeaderSize + ph->ImageSize);

		if ( (ph->ClutType & 0x3F) == TIM2_RGB16 ) m_iPalBytesPerPixel = 2;
		else if ( (ph->ClutType & 0x3F) == TIM2_RGB24 ) m_iPalBytesPerPixel = 3;
		else m_iPalBytesPerPixel = 4;

		iNbCols = 1<<m_iImageOriginalDepth;

		memset( m_pOriginalPalette, 255, 256 * 4 );

		int i;
		for ( i=0; i<256; i++ )
		{
			m_pPalette32[ i ] = (255) | (255<<8) | (0<<16) | (0<<24);
		}

		memcpy( m_pOriginalPalette, pClut, iNbCols * m_iPalBytesPerPixel );
	
		switch ( m_iPalBytesPerPixel )
		{
			case 2:
			{
				int i;
				for ( i=0; i<iNbCols; i++ )
				{
					unsigned int color = ((unsigned int*)m_pOriginalPalette)[ i ];
					unsigned int r, g, b, a = 255; 
				
					r = ((color >>  0) & 0x1f) << 3;
					g = ((color >>  5) & 0x1f) << 3;
					b = ((color >> 10) & 0x1f) << 3;
				
					m_pPalette32[ i ] = (a<<24) | (r<<16) | (g<<8) | (b);
				}
				break;
			}
			case 3:
			{
				union
				{
					unsigned int uiValue;
					unsigned char cValue[4];
				} color;
				int n;
				for ( n=0; n<iNbCols; n++ )
				{
					color.uiValue = *(unsigned int *)(m_pOriginalPalette+n*3);

					if(!bRedAsAlpha)
					{
						uint8 red = color.cValue[0];
						color.cValue[0] = color.cValue[2];
						color.cValue[2] = red;
			
						color.cValue[3] = 128;
						m_pPalette32[n] = color.uiValue;
					}
					else
					{
						color.cValue[3] = color.cValue[2] / 2;
						color.cValue[0] = 255;
						color.cValue[1] = 255;
						color.cValue[2] = 255;
						m_pPalette32[n] = color.uiValue;
					}
				}
				break;
			}
			case 4:
			{
				union
				{
					unsigned int uiValue;
					unsigned char cValue[4];
				} color;
				int n;
				for ( n=0; n<iNbCols; n++ )
				{
					color.uiValue = *(unsigned int *)(m_pOriginalPalette+n*4);

					if(!bRedAsAlpha)
					{
						uint8 red = color.cValue[0];
						color.cValue[0] = color.cValue[2];
						color.cValue[2] = red;
			
						color.cValue[3] = 128;
						m_pPalette32[n] = color.uiValue;
					}
					else
					{
						color.cValue[3] = color.cValue[2] / 2;
						color.cValue[0] = 255;
						color.cValue[1] = 255;
						color.cValue[2] = 255;
						m_pPalette32[n] = color.uiValue;
					}
				}
				break;
			}
		}

		m_iClutAddr = GIFStaticDMA.GetPointer();

		for(int i = 0; i < iNbCols; i++)
		{
			if(m_iImageOriginalDepth == 8)
				GIFStaticDMA.Add32(m_pPalette32[(i & 231) + ((i & 8) << 1) + ((i & 16) >> 1)]);
			else
				GIFStaticDMA.Add32(m_pPalette32[i]);
		}

		GIFStaticDMA.DMARet();
	}

	delete [] pTmpData;

	return true;
}

void CTexture::Upload(uint32 uGSPage)
{
	uint32 iClutPtr = ((uGSPage + 15) << 5);
	uint32 uQSizeClut = (256 >> 2);
	uint32 iDstPtr = (uGSPage << 5);

	if(m_iImageOriginalDepth == 8)
	{
		GIFDynamicDMA.Add128(GS_GIFTAG_BATCH(4, 0, 0, 0, 
							 GIF_FLG_PACKED, GS_BATCH_1(GIF_REG_A_D)));

		GIFDynamicDMA.Add64(BITBLTBUF_SET(0, 0, 0, iClutPtr, 1, 0));
		GIFDynamicDMA.Add64(BITBLTBUF);

		GIFDynamicDMA.Add64(TRXPOS_SET(0, 0, 0, 0, 0));
		GIFDynamicDMA.Add64(TRXPOS);

		GIFDynamicDMA.Add64(TRXREG_SET(16, 16));
		GIFDynamicDMA.Add64(TRXREG);

		GIFDynamicDMA.Add64(TRXDIR_SET(0));
		GIFDynamicDMA.Add64(TRXDIR);

		GIFDynamicDMA.Add128(GS_GIFTAG_BATCH(uQSizeClut, 0, 0, 0, 
							 GIF_FLG_IMAGE, 0));

		GIFDynamicDMA.DMACall(m_iClutAddr);
	}
        else {
          fprintf(stderr,"WARNING: n_iImageOriginalDepth != %d\n",
                  m_iImageOriginalDepth);
        }

	// TODO: Add support for 16 entry clut

	GIFDynamicDMA.Add128(GS_GIFTAG_BATCH(4, 0, 0, 0, 
						 GIF_FLG_PACKED, GS_BATCH_1(GIF_REG_A_D)));

	int TBW = m_iWidth / 64;

	GIFDynamicDMA.Add64(BITBLTBUF_SET(0, 0, 0, iDstPtr, (TBW == 0) ? 1 : TBW, TSMFromColourDepth(m_iImageDepth)));
	GIFDynamicDMA.Add64(BITBLTBUF);

	GIFDynamicDMA.Add64(TRXPOS_SET(0, 0, 0, 0, 0));
	GIFDynamicDMA.Add64(TRXPOS);

	GIFDynamicDMA.Add64(TRXREG_SET(m_iWidth, m_iHeight));
	GIFDynamicDMA.Add64(TRXREG);

	GIFDynamicDMA.Add64(TRXDIR_SET(0));
	GIFDynamicDMA.Add64(TRXDIR);

	uint32 iQWC = (m_iDataSize) >> 4;

	GIFDynamicDMA.Add128(GS_GIFTAG_BATCH(iQWC,1,0,0,GIF_FLG_IMAGE,0));
	GIFDynamicDMA.DMACall(m_iTexAddr);

#if 0
        /* Dummy tag to bring EOP=1 closer */
        GIFDynamicDMA.Add128(GS_GIFTAG_BATCH(1,1,0,0,GIF_FLG_REGLIST,
                             GS_BATCH_2(GIF_REG_RGBAQ,GIF_REG_RGBAQ)));
        GIFDynamicDMA.Add64(RGBAQ_SET(0x80,0x80,0x80,0x80,0));
        GIFDynamicDMA.Add64(RGBAQ_SET(0x80,0x80,0x80,0x80,0));
#endif

	TBW = (1 << m_l2Width) / 64;

	m_iTex0 = TEX0_SET(iDstPtr, (TBW == 0) ? 1 : TBW, TSMFromColourDepth(m_iImageOriginalDepth), m_l2Width, m_l2Height, 1, 0, iClutPtr, 0, 0, 0, 1);
}

void CTexture::Select() {
	VIFDynamicDMA.StartDirect();
	VIFDynamicDMA.StartAD();
	VIFDynamicDMA.AddAD(0, TEXFLUSH);
	VIFDynamicDMA.AddAD(ALPHA_SET(0, 1, 0, 1, 128), ALPHA_1);
	VIFDynamicDMA.AddAD(TEX1_SET(1, 0, 1, 1, 0, 0, 0), TEX1_1);
	VIFDynamicDMA.AddAD(m_iTex0, TEX0_1);
	VIFDynamicDMA.EndAD();
	VIFDynamicDMA.EndDirect();
}

int Tim2CheckFileHeader(void *pTim2)
{
	TIM2_FILEHEADER *pFileHdr = (TIM2_FILEHEADER *)pTim2;
	int i;

	if(pFileHdr->FileId[0]=='T' || pFileHdr->FileId[1]=='I' || pFileHdr->FileId[2]=='M' || pFileHdr->FileId[3]=='2') {
		i = 1;
	} 
	else if(pFileHdr->FileId[0]=='C' || pFileHdr->FileId[1]=='L' || pFileHdr->FileId[2]=='T' || pFileHdr->FileId[3]=='2') {
		i = 2;
	} 
	else 
	{
		printf("Tim2CheckFileHeaer: TIM2 is broken %02X,%02X,%02X,%02X\n",
					pFileHdr->FileId[0], pFileHdr->FileId[1], pFileHdr->FileId[2], pFileHdr->FileId[3]);
		return(0);
	}

	if(!(pFileHdr->FormatVersion==0x03 ||
		pFileHdr->FormatVersion==0xEE ||
	    (pFileHdr->FormatVersion==0x04 && (pFileHdr->FormatId==0x00 || pFileHdr->FormatId==0x01)))) {
		printf("Tim2CheckFileHeaer: TIM2 is broken (2)\n");
		return(0);
	}
	return(i);
}

TIM2_PICTUREHEADER *Tim2GetPictureHeader(void *pTim2, int imgno)
{
	TIM2_FILEHEADER *pFileHdr = (TIM2_FILEHEADER *)pTim2;
	TIM2_PICTUREHEADER *pPictHdr;
	int i;

	if(imgno>=pFileHdr->Pictures) {
		printf("Tim2GetPictureHeader: Illegal image no.(%d)\n", imgno);
		return(NULL);
	}

	if(pFileHdr->FormatId==0x00) {
		pPictHdr = (TIM2_PICTUREHEADER *)((char *)pTim2 + sizeof(TIM2_FILEHEADER));
	} else {
		pPictHdr = (TIM2_PICTUREHEADER *)((char *)pTim2 + 0x80);
	}

	for(i=0; i<imgno; i++) {
		pPictHdr = (TIM2_PICTUREHEADER *)((char *)pPictHdr + pPictHdr->TotalSize);
	}
	return(pPictHdr);
}
